/***************************************************************************\
  "lexer.hpp" - Lexical Analyzer (Phase B)
  Revision 15; 10 Nov 2002 - 15 March 2004
  Copyright (C) 2002-2004  Kevin P. Albrecht

  References:
    MCD 2.1.5; 2.1.5.1
\***************************************************************************/

#ifndef LEXER_HPP
#define LEXER_HPP

/////////////////////////
// DEFINES //////////////
/////////////////////////

#define LEXER_DIGITS "0123456789"
#define LEXER_NUMBER_ELEMENTS "0123456789."
#define LEXER_LETTERS "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ_"
#define LEXER_SPECIALS "+-*/\\=(),<>$.[]\":;{}&#^"
#define LEXER_MIN_WORD_LEN 2
#define LEXER_MAX_WORD_LEN 13

/////////////////////////
// INCLUDES /////////////
/////////////////////////

// Standard //

// Local //
#include "inputmanager.hpp"
#include "token.hpp"
#include "indentstack.hpp"

/////////////////////////
// DECLARATIONS /////////
/////////////////////////

class Lexer
{
  public:
    Lexer ();
    ~Lexer ();
    Token *get (Token *oldToken);

  private:
    IndentStack *mIndentStack;

    bool checkForInclude ();
    void skipBlankSpaces ();
    void checkIndent ();

    Token *getWord ();
    Token *getNumber ();
    Token *getString ();
    Token *getEOL ();
    Token *getSpecial ();
    Token *getEOF ();
    Token *getError ();
};

/////////////////////////
// DEFINITIONS //////////
/////////////////////////

Lexer::Lexer ()
{
  mIndentStack = new IndentStack ();
}

Lexer::~Lexer ()
{
  delete mIndentStack;
}

//// Get next token from buffer, delete the oldToken
Token *Lexer::get (Token *oldToken)
{
  char c;

  // This allows the current line to be incremented when this is
  //  the first token of a new line
  if (oldToken)
  {
    if (oldToken->code == codeEOL)
    {
      Global::nCurrentLine++;
    }
  }

  // Find indentation //
  if (oldToken)
  {
    if ((oldToken->code == codeEOL) ||
        (oldToken->code == codeStart)
       )
    {
      checkIndent ();
    }
  }

  // Delete the old token //
  if (oldToken) delete oldToken;

  skipBlankSpaces ();
  if (checkForInclude() == true)
  {
    checkIndent ();
  }
  skipBlankSpaces ();
  c = gInputManager->get ();

  if (k_setfind (c, LEXER_LETTERS))       return getWord ();
  else if (k_setfind (c, LEXER_DIGITS))   return getNumber ();
  else if (c == '\"')                     return getString ();
  else if (c == 0)                        return getEOL ();
  else if (k_setfind (c, LEXER_SPECIALS)) return getSpecial ();
  else if (c == INPUTMANAGER_EOF)         return getEOF ();
  else                                    return getError ();

  // This would allow ';' for EOLs:
  // else if (c == 0) || (c == ';'))    return getEOL ();
}

//// Check for inlusion symbol '$'; return true if found
bool Lexer::checkForInclude ()
{
  char *filename;
  int pos;

  if (gInputManager->get() == '$')
  {
    filename = k_new_s (256);
    k_copy (filename, "");
    pos = 0;

    gInputManager->seek ();
    while (gInputManager->get() != 0)
    {
      filename[pos++] = gInputManager->get();
      gInputManager->seek ();
    }

    filename[pos - 1] = '\0';

    gInputManager->include (filename);

    k_del_s (filename);

    return true;
  }

  return false;
}

//// Remove blank spaces from buffer; includes comments, and all whitespace
void Lexer::skipBlankSpaces ()
{
  int nLine;

  while (((gInputManager->get() <= ' ') && (gInputManager->get() > 0)) || (gInputManager->get() == '{'))
  {
    if (gInputManager->get() == '{')
    {
      gInputManager->seek ();
      if (gInputManager->get() == '{') // single line comment
      {
        while (gInputManager->get() != 0)
        {
          gInputManager->seek ();
        }
        if (gInputManager->get() == 0)
        {
          // DEBUGGED 10/29/02 //
          // By returning now, lex_get() can return an EOL token //
          return;
        }
      }
      else // multiple line comment
      {
        nLine = Global::nCurrentLine;
        while (gInputManager->get() != '}')
        {
          if (gInputManager->get() == INPUTMANAGER_EOF)
          {
            Global::nCurrentLine = nLine;
            err (true, EF_INCOMPLETE_COMMENT, NULL);
            errterm ();
          }
          else if (gInputManager->get() == 0) // this allows lines to be increased when EOL is passed
          {
            Global::nCurrentLine++;
          }
          gInputManager->seek ();
        }
      }
    }
    gInputManager->seek ();
  }
}

//// Check for indentation
void Lexer::checkIndent ()
{
  int nSpaces = 0, nLine = 0;

  while (((gInputManager->get() <= ' ') && (gInputManager->get() > 0)) || (gInputManager->get() == '{'))
  {
    if (gInputManager->get() == '{') // comment
    {
      nSpaces++;
      gInputManager->seek ();

      if (gInputManager->get() == '{') // single line comment
      {
        while (gInputManager->get() != 0)
        {
          nSpaces++;
          gInputManager->seek ();
        }
        if (gInputManager->get() == 0)
        {
          // break here since the end of line is reached //
          break;
        }
      }
      else // multiple line comment
      {
        nLine = Global::nCurrentLine;
        while (gInputManager->get() != '}')
        {
          if (gInputManager->get() == INPUTMANAGER_EOF)
          {
            Global::nCurrentLine = nLine;
            err (true, EF_INCOMPLETE_COMMENT, NULL);
            errterm ();
          }
          else if (gInputManager->get() == 0)
          {
            Global::nCurrentLine++; // this allows lines to be increased when EOL is passed
            nSpaces = 0; // this resets the number of spaces on this line
          }
          else
          {
            nSpaces++;
          }
          gInputManager->seek ();
        }
        nSpaces++;
        gInputManager->seek();
      }
    }
    else if (gInputManager->get() == '\t') // tab
    {
      nSpaces += GLOBAL_LEXER_TAB_WIDTH;
      gInputManager->seek ();
    }
    else // other whitespace
    {
      nSpaces++;
      gInputManager->seek ();
    }
  }

  // ensure the line is not empty //
  if (gInputManager->get() != 0)
  {
    if (nSpaces < Global::nIndent)
    {
      while ((mIndentStack->size() != 0) && (mIndentStack->top() != nSpaces))
      {
        gScopeManager->end ();
        mIndentStack->pop ();
      }
      if (mIndentStack->top () != nSpaces)
      {
        err (true, EF_INDENT_ERROR, NULL);
        errterm ();
      }
    }
    else if (nSpaces > Global::nIndent)
    {
      mIndentStack->push (nSpaces);
      gScopeManager->begin ();
    }
    Global::nIndent = nSpaces;
  }
}

//// Return a word from the buffer
Token *Lexer::getWord ()
{
  // Vars //
  Token *token = new Token();
  char c = gInputManager->get ();
  int nLexemePos = 0;
  int nLen;

  // Put the characters of the token into 'token->lexeme' //
  do {
    token->lexeme[nLexemePos++] = c;
    gInputManager->seek();
    c = gInputManager->get ();
  } while (k_setfind (c, LEXER_LETTERS) || k_setfind (c, LEXER_DIGITS));
  token->lexeme[nLexemePos] = '\0';

  // Length of the token //
  nLen = strlen (token->lexeme);

  // Assume it's an identifier //
  token->code = codeIdentifier;

  // Verify length of lexeme //
  if ((nLen >= LEXER_MIN_WORD_LEN) && (nLen <= LEXER_MAX_WORD_LEN))
  {
    switch (nLen)
    {
      case 2:
        if (strcmp (token->lexeme, "or") == 0) token->code = codeOR;
        break;
      case 3:
        if (strcmp (token->lexeme, "and") == 0) token->code = codeAND;
        else if (strcmp (token->lexeme, "def") == 0) token->code = codeDEF;
        else if (strcmp (token->lexeme, "int") == 0) token->code = codeINT;
        else if (strcmp (token->lexeme, "rat") == 0) token->code = codeRAT;
        else if (strcmp (token->lexeme, "str") == 0) token->code = codeSTR;
        else if (strcmp (token->lexeme, "pop") == 0) token->code = codePOP;
        else if (strcmp (token->lexeme, "mod") == 0) token->code = codeMOD;
        break;
      case 4:
        if (strcmp (token->lexeme, "file") == 0) token->code = codeFILE;
        else if (strcmp (token->lexeme, "goto") == 0) token->code = codeGOTO;
        else if (strcmp (token->lexeme, "cond") == 0) token->code = codeCOND;
        else if (strcmp (token->lexeme, "loop") == 0) token->code = codeLOOP;
        else if (strcmp (token->lexeme, "push") == 0) token->code = codePUSH;
        else if (strcmp (token->lexeme, "rout") == 0) token->code = codeROUT;
        else if (strcmp (token->lexeme, "bool") == 0) token->code = codeBOOL;
        else if (strcmp (token->lexeme, "true") == 0)
        {
          token->code = codeBoolLiteral;
          token->set (k_new_b (true));
        }
        break;
      case 5:
        if (strcmp (token->lexeme, "const") == 0) token->code = codeCONST;
        else if (strcmp (token->lexeme, "false") == 0)
        {
          token->code = codeBoolLiteral;
          token->set (k_new_b (false));
        }
        break;
      case 6:
        if (strcmp (token->lexeme, "return") == 0) token->code = codeRETURN;
        break;
      case 9:
        if (strcmp (token->lexeme, "_prim_cin") == 0) token->code = code_PRIM_CIN;
        else if (strcmp (token->lexeme, "_prim_asc") == 0) token->code = code_PRIM_ASC;
        break;
      case 10:
        if (strcmp (token->lexeme, "_prim_cout") == 0) token->code = code_PRIM_COUT;
        else if (strcmp (token->lexeme, "_prim_rand") == 0) token->code = code_PRIM_RAND;
        else if (strcmp (token->lexeme, "_prim_stop") == 0) token->code = code_PRIM_STOP;
        break;
      case 11:
        if (strcmp (token->lexeme, "_prim_fio_r") == 0) token->code = code_PRIM_FIO_R;
        else if (strcmp (token->lexeme, "_prim_fio_w") == 0) token->code = code_PRIM_FIO_W;
        else if (strcmp (token->lexeme, "_prim_fio_c") == 0) token->code = code_PRIM_FIO_C;
        break;
      case 12:
        if (strcmp (token->lexeme, "_prim_fio_or") == 0) token->code = code_PRIM_FIO_OR;
        else if (strcmp (token->lexeme, "_prim_fio_ow") == 0) token->code = code_PRIM_FIO_OW;
        break;
      case 13:
        if (strcmp (token->lexeme, "_prim_fio_eof") == 0) token->code = code_PRIM_FIO_EOF;
        break;
    }
  }

  /*REMOVED (10-18-02)
  // Mark type as identifier //
  if (token->code == codeIdentifier)
  {
    token->type = tIdentifier;
  }*/

  return token;
}

Token *Lexer::getNumber ()
{
  // Variables //
  Token *token = new Token();
  int nPos = 0;
  int nMantissaLen = -1; // Assume the token is an integer

  token->lexeme[nPos] = gInputManager->get ();
  gInputManager->seek ();

  while (k_setfind (gInputManager->get(), LEXER_NUMBER_ELEMENTS))
  {
    if (nMantissaLen > -1) // if currently in mantissa, increase mantissa length
    {
      nMantissaLen++;
    }
    if (gInputManager->get() == '.')
    {
      nMantissaLen++;
    }
    token->lexeme[++nPos] = gInputManager->get();
    gInputManager->seek ();
  }

  // Fill any empty mantissa places with zeros //
  if (nMantissaLen > -1)
  {
    while (nMantissaLen < GLOBAL_MANTISSA_SIZE)
    {
      nMantissaLen++;
      token->lexeme[++nPos] = '0';
    }
  }

  // End lexeme string //
  token->lexeme[++nPos] = 0;

  if (nMantissaLen > -1) // rational number
  {
    token->code = codeRatLiteral;
    token->set (k_new_d(k_conv_sd(token->lexeme)));
  }
  else // integer number
  {
    token->code = codeIntLiteral;
    token->set (k_new_n(k_conv_si(token->lexeme)));
  }

  return token;
}

Token *Lexer::getString ()
{
  // Variables //
  Token *token = new Token();
  char c;
  int pos = 0;

  // Token is a string literal //
  token->code = codeStrLiteral;

  // Opening quote //
  token->lexeme[pos++] = '\"';

  // Get the string //
  gInputManager->seek (); // skip opening quote
  c = gInputManager->get ();
  while (c != INPUTMANAGER_EOF)
  {
    if (c == '\"') // current char is quote
    {
      gInputManager->seek ();
      c = gInputManager->get ();
      if (c != '\"')
      {
        break; // not a double quote, so end the string
      }
    }
    else if (c == 0) // end of line reached
    {
      c = ' '; // replace end-line with blank
      err (true, EF_INCOMPLETE_STR_LIT, NULL);
      break;
    }

    // Append current char to string //
    token->lexeme[pos++] = c;

    // Get next char //
    gInputManager->seek ();
    c = gInputManager->get ();
  }

  if (c == INPUTMANAGER_EOF)
  {
    err (true, EF_UNEXPECTED_EOF, NULL);
  }

  token->lexeme[pos++] = '\"'; // closing quote
  token->lexeme[pos] = '\0';

  // After this, token->lexeme will contain the string with quotes and //
  // token->value will contain the string minus the quotes             //
  token->set (k_sub (token->lexeme, 1, strlen (token->lexeme) - 1));

  return token;
}

Token *Lexer::getEOL ()
{
  Token *token = new Token ();

  token->code = codeEOL;
  token->lexeme[0] = 0;

  gInputManager->seek ();

  return token;
}

Token *Lexer::getSpecial ()
{
  // Variables //
  Token *token = new Token ();
  char c;
  int pos = 0;

  c = gInputManager->get ();
  token->lexeme[pos++] = c;

  switch (c) {
    case '+': token->code = codePLUS;         gInputManager->seek (); break;
    case '-': token->code = codeMINUS;        gInputManager->seek (); break;
    case '*': token->code = codeTIMES;        gInputManager->seek (); break;
    case '/': token->code = codeFORWARDSLASH; gInputManager->seek (); break;
    case '\\': token->code = codeBACKSLASH;   gInputManager->seek (); break;
    case '=': token->code = codeEQUALS;       gInputManager->seek (); break;
    case '(': token->code = codeLEFTPAREN;    gInputManager->seek (); break;
    case ')': token->code = codeRIGHTPAREN;   gInputManager->seek (); break;
    case ',': token->code = codeCOMMA;        gInputManager->seek (); break;
    case '$': token->code = codeDOLLAR;       gInputManager->seek (); break;
    case '.': token->code = codeDOT;          gInputManager->seek (); break;
    case '[': token->code = codeLEFTBRACK;    gInputManager->seek (); break;
    case ']': token->code = codeRIGHTBRACK;   gInputManager->seek (); break;
    case '\"': token->code = codeQUOTEMARK;   gInputManager->seek (); break;
    case ';': token->code = codeSEMICOLON;    gInputManager->seek (); break;
    case ':': token->code = codeCOLON;        gInputManager->seek (); break;
    case '&': token->code = codeANDSIGN;      gInputManager->seek (); break;
    case '#': token->code = codePOUNDSIGN;    gInputManager->seek (); break;
    case '^': token->code = codeEXPONENT;     gInputManager->seek (); break;
    case '<': // < or <= or <> or <-
              gInputManager->seek ();
              c = gInputManager->get ();
              if (c == '=') // <=
              {
                token->lexeme[pos++] = '=';
                token->code = codeLTEQUAL;
                gInputManager->seek ();
              }
              else if (c == '>') // <>
              {
                token->lexeme[pos++] = '>';
                token->code = codeNOTEQUAL;
                gInputManager->seek ();
              }
              else if (c == '-') // <-
              {
                token->lexeme[pos++] = '-';
                token->code = codeARROW;
                gInputManager->seek ();
              }
              else // <
              {
                token->code = codeLT;
              }
              break;
    case '>': // > or >=
              gInputManager->seek ();
              c = gInputManager->get ();
              if (c == '=') // >=
              {
                token->lexeme[pos++] = '=';
                token->code = codeGTEQUAL;
                gInputManager->seek ();
              }
              else // >
              {
                token->code = codeGT;
              }
              break;
    default: // error
             token->code = codeNull;
             gInputManager->seek ();
             token->lexeme[pos] = '\0'; // must be here so that lexeme can be reported
             err (true, EF_UNDEFINED_TOKEN, token->lexeme);
             break;
  }
  token->lexeme[pos] = '\0';

  return token;
}

Token *Lexer::getEOF ()
{
  Token *token = new Token();

  token->code = codeEOF;
  token->lexeme[0] = 0;

  return token;
}

Token *Lexer::getError ()
{
  Token *token = new Token ();

  token->code = codeError;
  token->lexeme[0] = gInputManager->get ();
  token->lexeme[1] = '\0';

  gInputManager->seek ();
  err (true, EF_UNDEFINED_TOKEN, token->lexeme);

  return token;
}

#endif
