/***************************************************************************\
  "symbol.hpp" - Symbol (Phase C)
  Revision 15; 10 Nov 2002 - 15 March 2004
  Copyright (C) 2002-2004  Kevin P. Albrecht
\***************************************************************************/

#ifndef SYMBOL_HPP
#define SYMBOL_HPP

/////////////////////////
// DEFINES //////////////
/////////////////////////

// External Defines //
// GLOBAL_VALUE_STRING_SIZE

/////////////////////////
// INCLUDES /////////////
/////////////////////////

// Standard //

// Local //

/////////////////////////
// DECLARATIONS /////////
/////////////////////////

typedef enum
{
  tNull, tVoid, tRoutine, tLabel, tBool, tInt, tRat, tStr
} Type;

char *gvsTypeStrings[] =
{
  "<null>", "void", "routine", "label", "bool", "int", "rat", "str"
};

typedef enum
{
  repNull, repIdent, repLiteral
} RepresentationType;

char *gvsRepresentationTypeStrings[] =
{
  "<null>", "identifier", "literal"
};

class Symbol
{
  public:
    Symbol (char *aKey);
    ~Symbol ();

    Symbol *copy ();
    void saveValue (); // ensures that value is not deleted when symbol is
    void set (RepresentationType aRep, Type aType, bool aConstant, void *aValue);

    RepresentationType rep;
    Type type;
    bool constant;

    char *uid; // used to uniquely identify symbols in diff scopes

    void *value; // the value held in the symbol:
                 // TYPE         VALUE          MEANING
                 // ----         -----          -------
                 // tRoutine     Routine*       routine
                 // tLabel       int*           statement index
                 // tBool        bool*          boolean
                 // tInt         int*           integer
                 // tRat         double*        rational
                 // tStr         char*          string
                 // tNull        (not defined)  (other)

    Symbol *next;
    char *key;
};

class Parameter
{
  public:
    Parameter (Type aType);
    Type type;
};

class LocalVar
{
  public:
    Symbol *symbol;
    bool active; // true if active in scope, false if not
    LocalVar *next;
    LocalVar *prev;
};

class Routine
{
  public:
    Routine ();
    ~Routine ();

    void addLocalVar (Symbol *aSymbol);
    void removeLocalVar (Symbol *aSymbol);
    void setCurrLocalVarToTail ();
    void setCurrLocalVarToHead ();
    Symbol *getPrevLocalVar ();
    Symbol *getNextLocalVar ();

    void addParameter (Type aType);

    Type returnType;
    Symbol *location;
    bool returnedFrom;
    int numOfParameters; // number of parameters
    Parameter *parameters[GLOBAL_ROUTINE_MAX_PARAMETERS]; // an array of all parameters

  private:
    int mLocalVars;
    LocalVar *mLocalVarHead; // a doubly-linked list holding all variables declared
    LocalVar *mLocalVarTail; //  ... in this routine; used for recursion
    LocalVar *mLocalVarCurr;
};

/////////////////////////
// DEFINITIONS //////////
/////////////////////////

Symbol::Symbol (char *aKey)
{
  rep = repNull;
  type = tNull;
  constant = false;

  uid = NULL;

  value = NULL;
  next = NULL;
  key = k_clone (aKey);
}

Symbol::~Symbol ()
{
  if (value) delete value;
  if (uid) delete uid;
  k_del_s (key);
}

//// Sets the symbol's value to NULL so that it will not be deleted
void Symbol::saveValue ()
{
  value = NULL;
}

//// Makes a copy of the symbol and returns it. Does not copy
////   the "next" part.
Symbol *Symbol::copy ()
{
  Symbol* temp = new Symbol (key);

  temp->rep = rep;
  temp->type = type;
  temp->constant = constant;

  if (temp->type == tRoutine)
  {
    err (true, EF_CANNOT_COPY_ROUTINE_SYMBOL, NULL);
    errterm ();
  }
  else if ((temp->type == tLabel) || (temp->type == tInt))
  {
    temp->value = k_new_n (*(int *)value);
  }
  else if (temp->type == tBool)
  {
    temp->value = k_new_b (*(bool *)value);
  }
  else if (temp->type == tRat)
  {
    temp->value = k_new_d (*(double *)value);
  }
  else if (temp->type == tStr)
  {
    temp->value = k_clone ((char *)value);
  }
  else
  {
    temp->value = NULL;
  }

  return temp;
}

void Symbol::set (RepresentationType aRep, Type aType, bool aConstant, void *aValue)
{
  rep = aRep;
  type = aType;
  constant = aConstant;

  // given value is NULL, so set the value as the default for each type //
  if (aValue == NULL)
  {
    // determine the type of the declaration //
    if (aType == tNull) value = NULL;
    else if (aType == tLabel) value = k_new_n (0);
    else if (aType == tRoutine) value = new Routine ();
    else if (aType == tInt) value = k_new_n (0);
    else if (aType == tBool) value = k_new_b (false);
    else if (aType == tRat) value = k_new_d (0.0);
    else if (aType == tStr) value = k_clone ("");
  }
  else
  {
    if (aType == tNull) value = NULL;
    else if (aType == tLabel) value = k_new_n (*(int *)aValue);
    else if (aType == tRoutine)
    {
      err (true, EF_CANNOT_COPY_ROUTINE_SYMBOL, NULL);
      errterm ();
    }
    else if (aType == tInt) value = k_new_n (*(int *)aValue);
    else if (aType == tBool) value = k_new_b (*(bool *)aValue);
    else if (aType == tRat) value = k_new_d (*(double *)aValue);
    else if (aType == tStr) value = k_clone ((char*)aValue);
  }
}

Parameter::Parameter (Type aType)
{
  type = aType;
}

Routine::Routine ()
{
  returnType = tNull;
  location = NULL;
  returnedFrom = false;

  numOfParameters = 0;
  for (int i = 0; i < GLOBAL_ROUTINE_MAX_PARAMETERS; i++)
  {
    parameters[i] = NULL;
  }
  mLocalVars = 0;
  mLocalVarHead = NULL;
  mLocalVarTail = NULL;
  mLocalVarCurr = NULL;
}

Routine::~Routine ()
{
  LocalVar *temp;
   
  // delete all the Parameters //
  for (int i = 0; i < numOfParameters; i++)
  {
    delete (parameters[i]);
  }

  // delete all the LocalVars //
  while (mLocalVarHead != NULL)
  {
    temp = mLocalVarHead->next;
    delete mLocalVarHead;
    mLocalVarHead = temp;
  }
}

//// Add a local variable to the list of local variables declared
////   in the routine
void Routine::addLocalVar (Symbol *aSymbol)
{
  LocalVar *lv;

  // Ensure that the variable is not added more than once //
  lv = mLocalVarHead;
  while (lv != NULL)
  {
    if (lv->symbol == aSymbol)
    {
      lv->active = true;
      return; // the variable has already been added
    }
    lv = lv->next;
  }

  // Put the new LocalVar at the head of the list //
  lv = new LocalVar ();
  lv->symbol = aSymbol;
  lv->active = true;
  lv->next = mLocalVarHead;
  lv->prev = NULL;
  if (mLocalVarHead != NULL) mLocalVarHead->prev = lv;
  mLocalVarHead = lv;
  if (mLocalVarTail == NULL) mLocalVarTail = lv;
  mLocalVars++;
}

//// Sets the active tag in the local var to false if found; this
////  is only used by the temporary generator
void Routine::removeLocalVar (Symbol *aSymbol)
{
  LocalVar *lv = mLocalVarHead;

  while (lv != NULL)
  {
    if (lv->symbol == aSymbol)
    {
      lv->active = false;
      mLocalVars--;
      break;
    }
    lv = lv->next;
  }
}

void Routine::setCurrLocalVarToTail ()
{
  mLocalVarCurr = mLocalVarTail;
}

void Routine::setCurrLocalVarToHead ()
{
  mLocalVarCurr = mLocalVarHead;
}

Symbol *Routine::getPrevLocalVar ()
{
  Symbol *temp;

  while (mLocalVarCurr)
  {
    if (mLocalVarCurr->active == true)
    {
      temp = mLocalVarCurr->symbol;
      mLocalVarCurr = mLocalVarCurr->prev;
      return temp;
    }
    else
    {
      mLocalVarCurr = mLocalVarCurr->prev;
    }
  }
  
  return NULL;
}

Symbol *Routine::getNextLocalVar ()
{
  Symbol *temp;

  while (mLocalVarCurr)
  {
    if (mLocalVarCurr->active == true)
    {
      temp = mLocalVarCurr->symbol;
      mLocalVarCurr = mLocalVarCurr->next;
      return temp;
    }
    else
    {
      mLocalVarCurr = mLocalVarCurr->next;
    }
  }

  return NULL;
}

//// Add a parameter to the Routine, this is for argument passing
void Routine::addParameter (Type aType)
{
  parameters[numOfParameters] = new Parameter (aType);
  numOfParameters++;
}

#endif
